/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import java.io.File;

/**
   This version creates a single hierarchical (nested) Model.
<p>
   Compare with
      http://threejs.org/examples/#webgl_geometries
   or
      https://stemkoski.github.io/Three.js/Shapes.html
   or
      http://www.smartjava.org/ltjs/chapter-02/04-geometries.html
*/
public class Geometries_R9
{
   public static void main(String[] args)
   {
      // Create the Scene object that we shall render.
      Scene scene = new Scene();

      // Create a two-dimensional array of Models.
      Model[][] models = new Model[3][5];

      // row 0
      models[0][0] = new ObjSimpleModel(new File("../assets/great_rhombicosidodecahedron.obj"));
      ModelShading.setColor(models[0][0], Color.red);

      models[0][1] = new ConeFrustum(0.5, 1.0, 1.0, 10, 10);
      ModelShading.setColor(models[0][1], Color.orange);

      models[0][2] = new Box(1.0, 1.0, 1.0);
      ModelShading.setColor(models[0][2], Color.lightGray);

      models[0][3] = new Axes3D(1, 1, 1, Color.red, Color.green, Color.blue);

      models[0][4] = new Sphere(1.0, 30, 30);
      ModelShading.setColor(models[0][4], Color.lightGray);

      // row 1
      models[1][0] = new Cylinder(0.5, 1.0, 30, 30);
      ModelShading.setColor(models[1][0], Color.blue.brighter().brighter());

      models[1][1] = new ObjSimpleModel(new File("assets/horse.obj"));
      ModelShading.setColor(models[1][1], Color.pink);

      models[1][2] = new GRSModel(new File("assets/grs/vinci.grs"));
      ModelShading.setColor(models[1][2], Color.blue);

      models[1][3] = new Tetrahedron();
      ModelShading.setColor(models[1][3], Color.cyan);

      models[1][4] = new ObjSimpleModel(new File("assets/small_rhombicosidodecahedron.obj"));
      ModelShading.setColor(models[1][4], Color.magenta);

      // row 2
      models[2][0] = new TriangularPrism(1.0, 1.0, 10);
      ModelShading.setColor(models[2][0], Color.green.darker().darker());

      models[2][1] = new GRSModel(new File("assets/grs/bronto.grs"));
      ModelShading.setColor(models[2][1], Color.red);

      models[2][2] = new Torus(0.75, 0.25, 30, 30);
      ModelShading.setRandomColors(models[2][2]);

      models[2][3] = new Octahedron();
      ModelShading.setColor(models[2][3], Color.blue);

      models[2][4] = new Cone(0.5, 1.0, 30, 30);
      ModelShading.setColor(models[2][4], Color.yellow);

      // Create x, y and z axes
      Model xyzAxes = new Axes3D(6, -6, 6, 0, 7, -7,  Color.red);

      // Create a "top level" Model that is a horizontal coordinate plane.
      Model topLevel = new PanelXZ(-6, 6, -7, 7);
      ModelShading.setColor(topLevel, Color.darkGray);

      // Add all the models as nested models of the top level Model
      for (int i = 0; i < models.length; i++)
         for (int j = 0; j < models[i].length; j++)
         {
            topLevel.nestedModels.add(models[i][j]);
         }
      topLevel.nestedModels.add(xyzAxes);

      // Add the top level Model to the Scene.
      Position topLevel_p = new Position(topLevel);
      scene.addPosition(topLevel_p);

      // Set up the camera's view frustum.
      double right  = 2.0;
      double left   = -right;
      double top    = 1.0;
      double bottom = -top;
      double near   = 1.0;
      scene.camera.projPerspective(left, right, bottom, top, near);
/*
      double fov    = 90.0;
      double aspect = 2.0;
      double near   = 1.0;
      scene.camera.projPerspective(fov, aspect, near);
*/
      // Create a framebuffer to render our scene into.
      int vp_width  = 1200;
      int vp_height = 600;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);

      long startTime, stopTime;
      startTime = System.currentTimeMillis();
      for (int k = 0; k < 360; k++)
      {
         // Place the top level Position in front of the camera.
         topLevel_p.matrix = Matrix.translate(0, -3, -10);
         // Rotate it.
         topLevel_p.matrix.mult( Matrix.rotateY(k) );

         // Place each model in the xz-plane and
         // also rotate each model on its own axis.
         for (int i = 0; i < models.length; i++)
         {
            for (int j = 0; j < models[i].length; j++)
            {
               // Place the model where it belongs in the xz-plane
               // and rotate the model on its own axis.
               models[i][j].nestedMatrix = Matrix.translate(4-4*i, 0, 6-3*j)
                                            .times(Matrix.rotateX(3*k))
                                            .times(Matrix.rotateY(3*k));
            }
         }

         // Render
       //Renderer.doAntialiasing = true;
         fb.clearFB(Color.black);
         Pipeline.render(scene, fb.vp);
         fb.dumpFB2File(String.format("PPM_Geometries_R9_Frame%03d.ppm", k));
       //fb.dumpFB2File(String.format("PNG_Geometries_R9_Frame%03d.png", k), "png");
      }
      stopTime = System.currentTimeMillis();
      System.out.println("Wall-clock time: " + (stopTime - startTime));
   }
}
