/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**
   Draw an interactive robot arm with shoulder, elbow, wrist, and finger joints.
*/
public class RobotArm1
{
   private static double shoulderRotation = 0.0;
   private static double    elbowRotation = 0.0;
   private static double    wristRotation = 0.0;
   private static double   fingerRotation = 0.0;

   private static double shoulderLength = 0.4;
   private static double    elbowLength = 0.3;
   private static double    wristLength = 0.2;
   private static double   fingerLength = 0.1;

   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the s/S keys to rotate the arm at the shoulder.");
      System.out.println("Use the e/E keys to rotate the arm at the elbow.");
      System.out.println("Use the w/W keys to rotate the arm at the wrist.");
      System.out.println("Use the f/F keys to rotate the arm at the finger.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   public static void main(String[] args)
   {
      print_help_message();

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Create a Position object that will hold the robot arm.
      Position arm_p = new Position();

      // Add the Position object to the scene.
      scene.addPosition(arm_p);

      // Push the position away from where the camera is.
      arm_p.translate(0, 0, -1);

      /*
         Be sure to draw a picture of the (simple) tree that this code creates.
      */
      Model arm = new Model();
      arm.addVertex(new Vertex(0, 0, 0));
      arm.addVertex(new Vertex(shoulderLength, 0, 0));
      arm.addLineSegment(new LineSegment(0, 1));
      // Add the arm Model to the Scene's Position.
      arm_p.model = arm;

      Model elbow = new Model();
      elbow.addVertex(new Vertex(0, 0, 0));
      elbow.addVertex(new Vertex(elbowLength, 0, 0));
      elbow.addLineSegment(new LineSegment(0, 1));
      arm.nestedModels.add( elbow );

      Model wrist = new Model();
      wrist.addVertex(new Vertex(0, 0, 0));
      wrist.addVertex(new Vertex(wristLength, 0, 0));
      wrist.addLineSegment(new LineSegment(0, 1));
      elbow.nestedModels.add( wrist );

      Model finger = new Model();
      finger.addVertex(new Vertex(0, 0, 0));
      finger.addVertex(new Vertex(fingerLength, 0, 0));
      finger.addLineSegment(new LineSegment(0, 1));
      wrist.nestedModels.add( finger );

      arm.setColor( Color.blue );

      // Initialize the nested matrices for the sub models.
      elbow.nestedMatrix  = Matrix.translate(shoulderLength, 0, 0);
      wrist.nestedMatrix  = Matrix.translate(elbowLength,    0, 0);
      finger.nestedMatrix = Matrix.translate(wristLength,    0, 0);


      // Define initial dimensions for a FrameBuffer.
      int width  = 512;
      int height = 512;

      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions.
      @SuppressWarnings("serial")
      InteractiveFrame app = new InteractiveFrame("Renderer 9", width, height)
      {
         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
          //System.out.println( e );

            char c = e.getKeyChar();
            if ('h' == c)
            {
               print_help_message();
               return;
            }
            else if ('d' == c)
            {
               Pipeline.debug = ! Pipeline.debug;
             //Clip.debug = ! Clip.debug;
             //RasterizeAntialias.debug = ! RasterizeAntialias.debug;
            }
            else if ('a' == c)
            {
               Pipeline.doAntialiasing = ! Pipeline.doAntialiasing;
            }
            else if ('s' == c)
            {
               shoulderRotation += 2.0;
            }
            else if ('S' == c)
            {
               shoulderRotation -= 2.0;
            }
            else if ('e' == c)
            {
               elbowRotation += 2.0;
            }
            else if ('E' == c)
            {
               elbowRotation -= 2.0;
            }
            else if ('w' == c)
            {
               wristRotation += 2.0;
            }
            else if ('W' == c)
            {
               wristRotation -= 2.0;
            }
            else if ('f' == c)
            {
               fingerRotation += 2.0;
            }
            else if ('F' == c)
            {
               fingerRotation -= 2.0;
            }

            // Set the nested matrices for the sub models.
            arm.nestedMatrix = Matrix.rotateZ(shoulderRotation);

            elbow.nestedMatrix = Matrix.translate(shoulderLength, 0, 0)
                                .times(Matrix.rotateZ(elbowRotation));

            wrist.nestedMatrix = Matrix.translate(elbowLength, 0, 0)
                                .times(Matrix.rotateZ(wristRotation));

            finger.nestedMatrix = Matrix.translate(wristLength, 0, 0)
                                 .times(Matrix.rotateZ(fingerRotation));

            // Render again.
            FrameBuffer fb = this.fbp.getFrameBuffer();
            fb.clearFB(Color.green);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );

            // Get the new size of the FrameBufferPanel.
            int w = this.fbp.getWidth();
            int h = this.fbp.getHeight();

            // Create a new FrameBuffer that fits the new window size.
            FrameBuffer fb = new FrameBuffer(w, h);
            this.fbp.setFrameBuffer(fb);
            fb.clearFB(Color.green);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }
      };
   }
}
