/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This is a simple, nested (hierarchical) model
   made up of a triangle with a sphere attached
   to each vertex.
*/
class SimpleModel_1a extends Model
{
   public SimpleModel_1a()
   {
      super();

      // Add a single triangle to the geometry of this model.
      double sin2PIover3 = Math.sin(2*Math.PI/3);
      Vertex v0 = new Vertex( 1,        0,       0);
      Vertex v1 = new Vertex(-0.5,  sin2PIover3, 0);
      Vertex v2 = new Vertex(-0.5, -sin2PIover3, 0);
      addVertex(v0, v1, v2);
      addLineSegment(new LineSegment(0, 1),
                     new LineSegment(1, 2),
                     new LineSegment(2, 0));
      setColor(Color.black);

      // Create three nested Models.
      Model m1 = new Sphere(0.5, 10, 10);
      m1.setColor(Color.red);
      Model m2 = new Sphere(0.5, 10, 10);
      m2.setColor(Color.green);
      Model m3 = new Sphere(0.5, 10, 10);
      m3.setColor(Color.blue);

      // Put the three nested Models into this Model.
      nestedModels.add(m1);
      nestedModels.add(m2);
      nestedModels.add(m3);

      // Place the three sub-models at the
      // corners of the main model's triangle.
      m1.nestedMatrix = Matrix.translate( 1.5,   0,               0);
      m2.nestedMatrix = Matrix.translate(-0.75,  1.5*sin2PIover3, 0);
      m3.nestedMatrix = Matrix.translate(-0.75, -1.5*sin2PIover3, 0);
   }
}


/**
   This program uses the the hierarchical model
   defined just above.
<p>
   Here is a sketch of the scene graph for this example.
<pre>{@code
                 Scene
                   |
                   |
               Position
              /    |    \
             /     |     \
            /      |      \
      Matrix  SimpleModel  nested Positions
        RT    /     |   \
             /      |    \
        Matrix     tri    nested Models
           I             /      |      \
                        /       |       \
                       /        |        \
                      /         |         \
                 Model        Model        Model
                /   \         /   \        /   \
               /     \       /     \      /     \
         Matrix  sphere  Matrix  sphere  Matrix  sphere
           T               T               T
}</pre>
*/
public class NestedModels_v1a
{
   public static void main(String[] args)
   {
      // Create the Scene object that we shall render.
      Scene scene = new Scene();

      // Create the top level Position.
      Position top_p = new Position();

      // Add the top level Position to the Scene.
      scene.addPosition(top_p);

      // Create an instance of our hierarchical Model.
      Model mainModel = new SimpleModel_1a();
      // Add the Moddel to the top level Position
      top_p.model = mainModel;

      // Create a floor Model.
      Model floor = new PanelXZ(-4, 4, -4, 4);
      floor.setColor(Color.black);
      Position floor_p = new Position(floor);
      floor_p.translate(0, -4, 0);
      // Push this model away from where the camera is.
      floor_p.translate(0, 0, -5);
      // Add the floor to the Scene.
      scene.addPosition(floor_p);


      // Create a framebuffer to render our scene into.
      int vp_width  = 1024;
      int vp_height = 1024;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);

      for (int i = 0; i <= 72; i++)
      {
         // Translate and rotate the WHOLE model.
         top_p.matrix2Identity();
         // Push the whole model away from where the camera is.
         top_p.translate(0, 0, -5);
         // Rotate and translate the whole model.
         top_p.rotateZ(5*i);
         top_p.translate(2, 0, 0);
//         top_p.rotateY(5*i);
//         top_p.rotateX(5*i);

         // Render
         fb.clearFB(java.awt.Color.lightGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_NestedModels_v1a_Frame%02d.ppm", i));
      }
   }
}
