/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This is a simple, nested (hierarchical) model
   made up of a triangle with another triangle
   attached to each vertex.
*/
class SimpleModel_v2 extends Model
{
   public Model triangle;

   public SimpleModel_v2()
   {
      super();

      // Add a single triangle to the geometry of this model.
      double sin2PIover3 = Math.sin(2*Math.PI/3);
      Vertex v0 = new Vertex( 1,        0,       0);
      Vertex v1 = new Vertex(-0.5,  sin2PIover3, 0);
      Vertex v2 = new Vertex(-0.5, -sin2PIover3, 0);
      addVertex(v0, v1, v2);
      addLineSegment(new LineSegment(0, 1),
                     new LineSegment(1, 2),
                     new LineSegment(2, 0));
      setColor(Color.black);

      // Create three nested Models.
      Model m1 = new Model();
      Model m2 = new Model();
      Model m3 = new Model();

      // Put the three nested Models into this Model.
      nestedModels.add(m1);
      nestedModels.add(m2);
      nestedModels.add(m3);

      // Give each of these nested Models a triangle as
      // a (deeper) nested Model.
      triangle = new Model();
      triangle.addVertex(v0, v1, v2);
      triangle.addLineSegment(new LineSegment(0, 1),
                              new LineSegment(1, 2),
                              new LineSegment(2, 0));
      triangle.setColor(Color.red);
      m1.nestedModels.add(triangle);
      m2.nestedModels.add(triangle);
      m3.nestedModels.add(triangle);

      // Place the three sub-models at the
      // corners of the this model's triangle.
      m1.nestedMatrix = Matrix.translate(1.5, 0, 0);
      m2.nestedMatrix = Matrix.rotateZ(120)
                       .times(Matrix.translate(1.5, 0, 0));
      m3.nestedMatrix = Matrix.rotateZ(240)
                       .times(Matrix.translate(1.5, 0, 0));
   }
}


/**
   This program uses the the hierarchical model
   defined just above.
<p>
   Here is a sketch of the scene graph for this example.
<pre>{@code
                 Scene
                   |
                   |
               Position
              /    |    \
             /     |     \
            /      |      \
      Matrix  SimpleModel  nested Positions
       RT     /     |   \
             /      |    \
        Matrix     tri    nested Models
           I             /      |      \
                        /       |       \
                       /        |        \
                      /         |         \
                 Model        Model        Model
                /   \         /   \        /   \
               /     \       /     \      /     \
         Matrix  nested  Matrix  nested  Matrix  nested
           T     Models    TR    Models    TR    Models
                       \           |            /
                         \         |          /
                           \       |        /
                             \     |      /
                               \   |    /
                                 Model
                                /  |  \
                               /   |   \
                         Matrix   tri   nested
                           R            Models
}</pre>
*/
public class NestedModels_v2
{
   public static void main(String[] args)
   {
      // Create the Scene object that we shall render.
      Scene scene = new Scene();

      // Create the top level Position.
      Position top_p = new Position();

      // Add the top level Position to the Scene.
      scene.addPosition(top_p);

      // Create an instance of our hierarchical Model.
      SimpleModel_v2 mainModel = new SimpleModel_v2();
      // Add the Moddel to the top level Position
      top_p.model = mainModel;

      // Create a floor Model.
      Model floor = new PanelXZ(-4, 4, -4, 4);
      floor.setColor(Color.black);
      Position floor_p = new Position(floor);
      floor_p.translate(0, -4, 0);
      // Push this model away from where the camera is.
      floor_p.translate(0, 0, -5);
      // Add the floor to the Scene.
      scene.addPosition(floor_p);


      // Create a framebuffer to render our scene into.
      int vp_width  = 1024;
      int vp_height = 1024;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);

      for (int i = 0; i <= 72; i++)
      {
         // Rotate the triangles WITHIN the model.
         mainModel.triangle.nestedMatrix = mainModel.triangle.nestedMatrix.times(Matrix.rotateX(5));

         // Rotate just one triangle WITHIN the scene.
       //mainModel.nestedModels.get(0).nestedMatrix = mainModel.nestedModels.get(0).nestedMatrix.times(Matrix.rotateX(5));

         // Translate and rotate the WHOLE model.
         top_p.matrix2Identity();
         // Push the whole model away from where the camera is.
         top_p.translate(0, 0, -5);
         // Rotate and translate the whole model.
         top_p.rotateZ(5*i);
         top_p.translate(2, 0, 0);
//         top_p.rotateY(5*i);
//         top_p.rotateX(5*i);

         // Render
         fb.clearFB(java.awt.Color.lightGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_NestedModels_v2_Frame%02d.ppm", i));
      }
   }
}
