/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This is a nested (hierarchical) model made
   up of four prisms connected into a ring.
<p>
   Here is a sketch of this model's scene graph.
<pre>{@code
            Prisms Model
            /    |     \
           /     |      \
     Matrix   geometry   nested Models
       I      (empty)      / |  | \
                          /  |  |  \
                        m1  m2  m3  m4
                          \  |  |  /
                           \ |  | /
                        TriangularPrism
}</pre>
*/
class Prisms extends Model
{
   public Prisms()
   {
      super();

      final double sqrt3 = Math.sqrt(3.0);

      Model m1 = new Model();
      Model m2 = new Model();
      Model m3 = new Model();
      Model m4 = new Model();

      this.nestedModels.add(m1);
      this.nestedModels.add(m2);
      this.nestedModels.add(m3);
      this.nestedModels.add(m4);

      // Create a TriangularPrism object.
      Model m5 = new TriangularPrism(1.0/sqrt3, 2.0, Math.PI/4.0, 25);

      // right
      m1.nestedModels.add(m5);
      m1.nestedMatrix = Matrix.translate(2+0.5/sqrt3, 0, 0);
      // left
      m2.nestedModels.add(m5);
      m2.nestedMatrix = Matrix.translate(-2-0.5/sqrt3, 0, 0)
                       .times(Matrix.rotateZ(180));
      // top
      m3.nestedModels.add(m5);
      m3.nestedMatrix = Matrix.rotateZ(90)
                       .times(Matrix.translate(2+0.5/sqrt3, 0, 0));
      // bottom
      m4.nestedModels.add(m5);
      m4.nestedMatrix = Matrix.rotateZ(-90)
                       .times(Matrix.translate(2+0.5/sqrt3, 0, 0));
   }
}


/**
   Here is a sketch of this program's scene graph.
   Only the Prisml node holds any geometry. All of
   the other nodes hold only a matrix.
<pre>{@code
                 Scene
                   |
                   |
               Position
              /    |    \
             /     |     \
       Matrix   Prisms    nested Positions
        RT     /  |   \
              /   |    \
        Matrix  empty   nested Models
          I               / |  | \
                         /  |  |  \
                       m1  m2  m3  m4
                         \  |  |  /
                          \ |  | /
                      TriangularPrism
}</pre>
*/
public class NestedPrisms_v1
{
   public static void main(String[] args)
   {
      // Create the Scene object that we shall render.
      Scene scene = new Scene();

      // Create the top level Position.
      Position top_p = new Position();

      // Add the top level Position to the Scene.
      scene.addPosition(top_p);

      Model prisms = new Prisms();
      prisms.setColor(Color.red);
      top_p.model = prisms;

      // Create a floor Model.
      Model floor = new PanelXZ(-4, 4, -4, 4);
      floor.setColor(Color.black);
      Position floor_p = new Position(floor);
      floor_p.translate(0, -4, 0);
      // Push this model away from where the camera is.
      floor_p.translate(0, 0, -5);
      // Add the floor to the Scene.
      scene.addPosition(floor_p);


      // Create a framebuffer to render our scene into.
      int vp_width  = 1024;
      int vp_height = 1024;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);

      // Spin the model 360 degrees arond two axes.
      for (int i = 0; i <= 180; i++)
      {
         top_p.matrix2Identity();
         // Push the model away from where the camera is.
         top_p.translate(0, 0, -8);
         top_p.rotateX(2*i);
         top_p.rotateY(2*i);
         top_p.rotateZ(2*i);

         // Render again.
         fb.clearFB(java.awt.Color.lightGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_NestedPrism_v1_Frame_%02d.ppm", i));
      }
   }
}
