/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**
   Draw an animation of a solar system with a sun, planet, and moon.
<p>
   In this version, the orbit of the planet is independent of the
   rotation of the sun, and the orbit of the moon is independent
   of the rotation of the planet. This version has the exact same
   scene graph structure as v2a, but this version uses different matrices.
<pre>{@code
              Scene
                |
                |
             Position
            /   |     \
           /    |      \
     Matrix   Model     nested Positions
       T     /  |  \                   \
            /   |   \                   \
      Matrix  empty  \                  empty
        R             \
                       nested Models
                        /           \
                       /             \
                   Model             Model
                  /  |              /  |  \
                 /   |             /   |   \
           Matrix   sun      Matrix  empty  \
             R                 T             \
                                            nested Models
                                             /          \
                                            /            \
                                        Model            Model
                                       /  |             /   |
                                      /   |            /    |
                                Matrix  planet    Matrix   moon
                                  R                 TR
}</pre>
*/
public class SolarSystem_v2b
{
   private static int fps = 20;

   private static double planetOrbitRadius = 5.0;
   private static double   moonOrbitRadius = 1.0;

   private static double planetOrbitRot = 0.0;
   private static double   moonOrbitRot = 0.0;

   private static double    sunAxisRot = 0.0;
   private static double planetAxisRot = 0.0;
   private static double   moonAxisRot = 0.0;

   private static double ecliptic = 7.0; // angle of the ecliptic plane

   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the f/F keys to slow down or speed up the frame rate.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the e/E keys to change the angle of the ecliptic plane.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }

   public static void main(String[] args)
   {
      print_help_message();

      /*
         See the above picture of the tree that this code creates.
      */

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Create a Position object that will hold the solar system.
      Position solarSystem_p = new Position();

      // Add the Position object to the scene.
      scene.addPosition(solarSystem_p);

      // Push the position away from where the camera is.
      solarSystem_p.translate(0, 0, -8);


      // Create the Model that holds the whole solar system.
      Model solarSystem = new Model();  // A Solar System
      // Add the solar system model to the Scene's Position.
      solarSystem_p.model = solarSystem;

      // Create the sun.
      Model sun = new Sphere(1.0, 10, 10);
      sun.setColor( Color.yellow );
      // Add the sun Model to the solar system Model.
      solarSystem.nestedModels.add(sun);

      // Create the Model that holds the planet-moon system.
      Model planetMoon = new Model();
      // Add the planet-moon Model to the solar system Model.
      solarSystem.nestedModels.add(planetMoon);

      // Create the planet.
      Model planet = new Sphere(0.5, 10, 10);
      planet.setColor( Color.blue );
      // Add the planet Model to the planet-moon Model.
      planetMoon.nestedModels.add(planet);

      // Create the moon.
      Model moon = new Sphere(0.2, 10, 10);
      moon.setColor( Color.green );
      // Add the moon Model to the planet-moon Model.
      planetMoon.nestedModels.add(moon);


      // Define initial dimensions for a FrameBuffer.
      int width  = 512;
      int height = 512;

      // Create an AnimationFrame containing a FrameBuffer
      // with the given dimensions.
      @SuppressWarnings("serial")
      InteractiveFrame app = new AnimationFrame("Renderer 9", width, height, fps)
      {
         // Implement the ActionListener interface.
         @Override public void actionPerformed(ActionEvent e)
         {
            //System.out.println( e );

            // Rotate the plane of the ecliptic
            // (rotate the solar system model's xz-plane about the x-axis).
            solarSystem.nestedMatrix = Matrix.rotateX(ecliptic);

            // Rotate the sun on it axis.
            sun.nestedMatrix = Matrix.rotateY(sunAxisRot);

            // Place the planet-moon in orbit around the sun.
            planetMoon.nestedMatrix = Matrix.translate(
                   planetOrbitRadius * Math.sin(planetOrbitRot * Math.PI/180),
                   0,
                   planetOrbitRadius * Math.cos(planetOrbitRot * Math.PI/180));

            // Rotate the planet on it axis.
            planet.nestedMatrix = Matrix.rotateY(planetAxisRot);

            // Place the moon in orbit around the planet and rotate the moon on its axis.
            moon.nestedMatrix = Matrix.translate(
                   moonOrbitRadius * Math.sin(moonOrbitRot * Math.PI/180),
                   0,
                   moonOrbitRadius * Math.cos(moonOrbitRot * Math.PI/180))
                 .times(Matrix.rotateY(moonAxisRot));

            // Update the parameters for the next frame.
            sunAxisRot -= 10.0;
            planetOrbitRot += 1.0;
            planetAxisRot -= 5.0;
            moonOrbitRot += 5.0;
            moonAxisRot += 10.0;

            // Render again.
            FrameBuffer fb = this.fbp.getFrameBuffer();
            fb.clearFB(Color.black);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }

         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
            //System.out.println( e );

            char c = e.getKeyChar();
            if ('h' == c)
            {
               print_help_message();
               return;
            }
            else if ('d' == c)
            {
               Pipeline.debug = ! Pipeline.debug;
             //Clip.debug = ! Clip.debug;
             //RasterizeAntialias.debug = ! RasterizeAntialias.debug;
            }
            else if ('f' == c)
            {
               fps -= 1;
               if (0 > fps) fps = 0;
               System.out.println("fps = " + fps);
            }
            else if ('F' == c)
            {
               fps += 1;
               System.out.println("fps = " + fps);
            }
            else if ('p' == c)
            {
               scene.camera.perspective = ! scene.camera.perspective;
            }
            else if ('e' == c)
            {
               ecliptic -= 1;
               if (0 == fps) fps = 1;
               System.out.println("ecliptic = " + ecliptic);
            }
            else if ('E' == c)
            {
               ecliptic += 1;
               System.out.println("ecliptic = " + ecliptic);
            }

            setFPS(fps);

            // Set up the camera's view volume.
            if (scene.camera.perspective)
            {
               scene.camera.projPerspective(-1, 1, -1, 1, 1);
            }
            else
            {
               scene.camera.projOrtho(-6, 6, -6, 6);
            }
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );

            // Get the new size of the FrameBufferPanel.
            int w = this.fbp.getWidth();
            int h = this.fbp.getHeight();

            // Create a new FrameBuffer that fits the new window size.
            FrameBuffer fb = new FrameBuffer(w, h);
            this.fbp.setFrameBuffer(fb);
            fb.clearFB(Color.black);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }
      };
   }
}
